/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse;

import static com.google.common.collect.Maps.asMap;
import static com.google.inject.Guice.createInjector;
import static org.fortiss.af3.exploration.alg.service.ExplorationServiceManager.getService;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.copy;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.disableNotificationsOf;

import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.function.Consumer;
import java.util.function.Predicate;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.swt.widgets.Display;
import org.fortiss.af3.exploration.alg.dse.backend.IExplorerBackend;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.Opt4JExplorerBackend;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.DseInjectorProvider;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.create.CreateInjectorProvider;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.create.CreateInterceptModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.create.CreateInterceptor;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.DecodeInjectorProvider;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.DecodeInterceptor;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.DecodingInterceptModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.evaluate.EvaluatorInjectorProvider;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.thread.MultiThreadingModule;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.exception.ExplorationException;
import org.fortiss.af3.exploration.alg.plot.XYPlotter;
import org.fortiss.af3.exploration.alg.service.IExpSpecTransformationService;
import org.fortiss.af3.exploration.model.ExplorationSpecification;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.model.solutions.ExplorationSolution;
import org.fortiss.af3.exploration.model.solutions.SingleExplorationSolution;
import org.fortiss.af3.exploration.moea.model.DseSpecification;
import org.fortiss.af3.exploration.moea.model.parameters.EAOptimizer;
import org.fortiss.af3.exploration.moea.model.parameters.IOptimizer;
import org.fortiss.af3.exploration.util.ExplorationUtils;
import org.jfree.ui.RefineryUtilities;

import com.google.inject.Guice;
import com.google.inject.Injector;

/**
 * Explorer - entry point to architectural exploration
 * 
 * @author huang
 */
public class Explorer {

	// TODO: Remove after complete migration to AF3 visualization
	/** Last {@link XYPlotter} (in order to close it before creating a new one. */
	protected XYPlotter lastPlotter;

	/** Maintains a reference to the original input set of {@link ExplorationTarget}s. */
	private Collection<ExplorationTarget<?>> copiedOrigTargets = new HashSet<>();

	/**
	 * Performs the DSE defined a given {@link DseSpecification} specification on a given
	 * system model (represented by a {@link SystemModelAdapter}).
	 * 
	 * @throws Exception
	 */
	public ExplorationSolution explore(DseSpecification dse,
			SystemModelAdapter<?, ?, ?, ?, ?, ?, ?> systemModelAdapter,
			Set<Class<? extends Phenotype>> requestedSolutions, boolean generateImplicitConstraints)
			throws Exception {
		return explore(dse, systemModelAdapter, requestedSolutions, null,
				generateImplicitConstraints);
	}

	/**
	 * Performs the DSE defined a given {@link DseSpecification} specification on a given
	 * system model (represented by a {@link SystemModelAdapter}).
	 */
	public ExplorationSolution explore(DseSpecification dseSpec,
			SystemModelAdapter<?, ?, ?, ?, ?, ?, ?> systemModelAdapter,
			Set<Class<? extends Phenotype>> requestedSolutions, IProgressMonitor progressMonitor,
			boolean generateImplicitConstraints) throws Exception {
		ExplorationSpecification origExpSpec = dseSpec.getTargetSpecification();

		ExplorationSpecification expSpec = perprocessExpSpec(dseSpec, systemModelAdapter,
				requestedSolutions, generateImplicitConstraints, origExpSpec);
		InternalSpecificationProvider specProvider =
				new InternalSpecificationProvider(expSpec, requestedSolutions);

		// Invoke the backend in order to perform the DSE
		IExplorerBackend<? extends IOptimizer> explorerBackend = createExplorerBackend(dseSpec,
				expSpec, systemModelAdapter, specProvider, requestedSolutions, progressMonitor);
		ExplorationSolution internalExplorationResult = explorerBackend.explore();
		conditionallyShowDebugVisualization(internalExplorationResult);

		// We use the original ExplorationSpecification for the returned result.
		ExplorationSolution explorationResult = copy(internalExplorationResult);
		explorationResult.setExplorationSpec(origExpSpec);

		// Remove the implicit ExplorationTargets from the metrics of each single solution.
		removeInternalTargets(explorationResult);

		return explorationResult;
	}

	/**
	 * Shows a visualization of the {@link ExplorationSolution} that includes the evaluations of
	 * internal constraints and objectives.
	 */
	private void conditionallyShowDebugVisualization(ExplorationSolution explorationSolution) {
		if(ExplorationUtils.isDebugVerboseEnabled()) {
			Display.getDefault().asyncExec(() -> plotResult(explorationSolution));
		}
	}

	/**
	 * Copies the input {@link ExplorationException} and applies the D3SE input transformation
	 * service to the copy.
	 * 
	 * @param dseSpec
	 *            input {@link DseSpecification} defining the parameters of the MOEA backend.
	 * @param systemModelAdapter
	 *            Representation of the input models.
	 * @param requestedSolutions
	 *            Set of user-requested artifacts produced by the DSE.
	 * @param generateImplicitConstraints
	 *            flag whether or not to generate some of the implicit constraints (debug).
	 * @param origExpSpec
	 *            original input {@link ExplorationSpecification}.
	 * @return copied and modified input {@link ExplorationSpecification}.
	 * @throws ExplorationException
	 *             if some of the applied transformation fails.
	 */
	protected ExplorationSpecification perprocessExpSpec(DseSpecification dseSpec,
			SystemModelAdapter<?, ?, ?, ?, ?, ?, ?> systemModelAdapter,
			Set<Class<? extends Phenotype>> requestedSolutions, boolean generateImplicitConstraints,
			ExplorationSpecification origExpSpec) throws ExplorationException {
		// Copy the input specification and prepare it for further processing.
		ExplorationSpecification expSpec = copyAndDisconnectExpSpec(origExpSpec);
		Injector inputInjector = createInjector(systemModelAdapter);

		// Execute the Constraint Transformation.
		List<ExplorationTarget<?>> targets = expSpec.getTargets();
		IExpSpecTransformationService expSpecSrv = getService(IExpSpecTransformationService.class);
		Map<ExplorationTarget<?>, ExplorationTarget<?>> inputTargets =
				asMap(new HashSet<>(targets), k -> null);
		Map<ExplorationTarget<?>, ExplorationTarget<?>> transformedTargets =
				expSpecSrv.transform(inputInjector, inputTargets);
		// @CodeFormatterOff
		transformedTargets.entrySet().stream().
			filter(e -> e.getValue() != null).
			forEach(e -> {
				targets.remove(e.getKey());
				targets.add(e.getValue());
		});
		// @CodeFormatterOn

		// Add implicit Constraints.
		if(generateImplicitConstraints) {
			expSpec.getTargets().addAll(ImplicitExplorationTargetFactory
					.createImplicitExplorationTargets(dseSpec, systemModelAdapter));
		}
		dseSpec.setTargetSpecification(expSpec);

		return expSpec;
	}

	/**
	 * Returns a copy of the given {@link ExplorationSpecification} and its referenced
	 * {@link ExplorationTarget}s. Additionally, the EMF notifications are disabled for these
	 * elements to ease manipulation of the data structure.
	 * 
	 * @param origExpSpec
	 *            {@link ExplorationSpecification} to be copied.
	 * @return Copy of the input {@link ExplorationSpecification} with disabled notifications.
	 */
	private ExplorationSpecification
			copyAndDisconnectExpSpec(ExplorationSpecification origExpSpec) {
		// Copy.
		ExplorationSpecification expSpec = copy(origExpSpec);
		expSpec.getTargets().clear();
		origExpSpec.getTargets().forEach(t -> {
			ExplorationTarget<?> expTarget = copy(t);
			copiedOrigTargets.add(expTarget);
			expSpec.getTargets().add(expTarget);
		});
		// Disable notifications on copied elements.
		disableNotificationsOf(expSpec);
		expSpec.getTargets().forEach(t -> disableNotificationsOf(t));

		return expSpec;
	}

	/** Helper method to issue an exception for invalid DSE configurations */
	private void createDesignSpaceExplorationConfigurationException(final String msg)
			throws Exception {
		throw new Exception("DSE configuration: " + msg);
	}

	/**
	 * Removes the implicit {@link ExplorationTarget}s that were added internally. They shall not be
	 * visible to the user.
	 */
	private void removeInternalTargets(ExplorationSolution explorationSolution) {
		Predicate<? super ExplorationTarget<?>> isOrigTarget = t -> !copiedOrigTargets.contains(t);
		Consumer<SingleExplorationSolution> removeNonOrigTargets =
				s -> s.keySet().stream().filter(isOrigTarget).forEach(t -> s.removeKey(t));
		explorationSolution.getSolutions().forEach(removeNonOrigTargets);
	}

	/** Explorer backend */
	private IExplorerBackend<? extends IOptimizer> createExplorerBackend(DseSpecification dse,
			ExplorationSpecification expSpec,
			SystemModelAdapter<?, ?, ?, ?, ?, ?, ?> systemModelAdapter,
			InternalSpecificationProvider specProvider,
			Set<Class<? extends Phenotype>> requestedSolutions, IProgressMonitor progressMonitor)
			throws Exception {

		DseInjectorProvider provider = new DseInjectorProvider();
		Injector dseInjector = Guice.createInjector(systemModelAdapter, specProvider,
				new CreateInterceptModule(new CreateInterceptor()), new CreateInjectorProvider(),
				new DecodingInterceptModule(new DecodeInterceptor()), new DecodeInjectorProvider(),
				new EvaluatorInjectorProvider(), new MultiThreadingModule(), provider);
		provider.setInjector(dseInjector);

		if(dse.getOptimizer() instanceof EAOptimizer) {
			return new Opt4JExplorerBackend(dse, expSpec, dseInjector, requestedSolutions,
					progressMonitor);
		}

		createDesignSpaceExplorationConfigurationException("Unsupported exploration backend.");

		// Never here
		return null;
	}

	/**
	 * Creates and launches the Dialog that displays the results from the exploration.
	 * Pareto-optimal points are displayed in a 2D-graph, while schedules can be inspected for
	 * each solution.
	 */
	public void plotResult(ExplorationSolution solutionSet) {
		if(solutionSet != null) {
			// Keep only one DSE result window
			if(lastPlotter != null) {
				lastPlotter.dispose();
			}

			// draw figure
			XYPlotter plotter = new XYPlotter(solutionSet,
					"Design Space Exploration Result (generations: N/A)");
			plotter.createPlot();
			plotter.pack();
			RefineryUtilities.centerFrameOnScreen(plotter);
			plotter.setVisible(true);

			lastPlotter = plotter;
		}
	}
}
