/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode;

import static org.fortiss.tooling.common.util.LambdaUtils.getFirst;

import java.util.Collection;
import java.util.Set;

import org.aopalliance.intercept.MethodInterceptor;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.Opt4JDseTaskProvider;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.DseInjectorProvider;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.EncodingProviderModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.evaluate.ValueProvider;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.genotype.ProvidingGenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.PhenotypeProviderGenerator;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.ProvidingPhenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.thread.ThreadedMethod;
import org.fortiss.af3.exploration.alg.dse.evaluator.BooleanValue;
import org.fortiss.af3.exploration.alg.dse.module.CreatorModule;
import org.fortiss.af3.exploration.alg.dse.module.DecoderModule;
import org.opt4j.core.Genotype;
import org.opt4j.core.Value;
import org.opt4j.core.problem.Decoder;
import org.opt4j.core.start.Opt4JTask;

import com.google.common.collect.ClassToInstanceMap;
import com.google.common.collect.MutableClassToInstanceMap;
import com.google.inject.AbstractModule;
import com.google.inject.Inject;
import com.google.inject.Injector;

/**
 * {@link Decoder} that handles dependencies between {@link Decoder}s, {@link Genotype}s, and
 * {@link Phenotype}s using the Guice framework.
 * 
 * @author diewald
 */
public class GuiceDecoder extends AbstractModule
		implements Decoder<ProvidingGenotype, ProvidingPhenotype> {

	/** Defines the desired exploration result type. */
	// TODO: Remove this hard coded line and get the desired result from the backend interface.
	protected Set<Class<? extends Phenotype>> resultTypes;

	/** Set of {@link CreatorModule}s provided by the enabled exploration features. */
	private Collection<Class<? extends DecoderModule<?>>> enabledDecoders;

	/** {@link Opt4JTask} provider used to obtain the opt4j {@link Injector}. */
	@Inject
	protected Opt4JDseTaskProvider opt4JProvider;

	/** Child {@link Injector} for the dependency management of the decode phase. */
	protected Injector decodingInjector;

	/** References the top-most injector of the DSE. */
	private Injector dseInjector;

	/**
	 * Holds the {@link MethodInterceptor} that is executed when a decode method capable of
	 * launching sub-decoders is hit.
	 */
	protected DecodeInterceptor decodingInterceptor;

	/** Constructor. Remembers the demanded {@link Phenotype}ss to be produced. */
	@Inject
	public GuiceDecoder(Set<Class<? extends Phenotype>> resultTypes,
			DseInjectorProvider dseInjectorProvider) {
		this.resultTypes = resultTypes;
		this.dseInjector = dseInjectorProvider.getInjector();
	}

	// TODO: Pass/calculate the set of enabled decoders with the constructor.
	/**
	 * Defines the decoders that are required in this DSE execution to produce the demanded results.
	 */
	public void setEnabledDecoders(Collection<Class<? extends DecoderModule<?>>> enabledDecoders) {
		this.enabledDecoders = enabledDecoders;
	}

	/** {@inheritDoc} */
	@Override
	@ThreadedMethod
	public ProvidingPhenotype decode(ProvidingGenotype genotype) {
		Injector decodingInjector = createDecodingInjector(genotype);
		DecodeInjectorProvider injProv = decodingInjector.getInstance(DecodeInjectorProvider.class);
		injProv.setInjector(decodingInjector);

		DecodeInterceptor interceptor = decodingInjector.getInstance(DecodeInterceptor.class);
		interceptor.setInterceptParameters(dseInjector, DecodeInjectorProvider.class,
				Phenotype.class, PhenotypeCache.class);

		// FIXME: Hack to get the correct instance: Is multiprovisioning supported by Guice?
		PhenotypeCache aggregatePhenotypes = decodingInjector.getInstance(PhenotypeCache.class);
		aggregatePhenotypes.addAllPreservedPhenos(genotype);
		Class<? extends Phenotype> resultType = getFirst(resultTypes).get();
		decodingInjector.getInstance(resultType);
		// The decoding Injector creates the Phenotype instances. They are cached in the
		// interceptor after each of the decode methods returns.

		// TODO(AD): Extract to separate method.
		PhenotypeProviderGenerator phenoProviderGenerator =
				decodingInjector.getInstance(PhenotypeProviderGenerator.class);
		EncodingProviderModule<Phenotype> phenotProviderModule =
				phenoProviderGenerator.generateInstanceProviders(aggregatePhenotypes.values());

		ClassToInstanceMap<Phenotype> solutionHolder = MutableClassToInstanceMap.create();
		solutionHolder.putAll(aggregatePhenotypes);
		ProvidingPhenotype providingPhenotype = new ProvidingPhenotype(genotype.getProviderModule(),
				phenotProviderModule, solutionHolder, genotype);

		return providingPhenotype;
	}

	/**
	 * Creates a decoding {@link Injector} as child {@link Injector} if the DSE-wide
	 * {@link Injector}. Thread-safe.
	 */
	private synchronized Injector createDecodingInjector(ProvidingGenotype genotype) {
		return opt4JProvider.getInjector().createChildInjector(this, genotype.getProviderModule());
	}

	/** {@inheritDoc} */
	@Override
	protected void configure() {
		install(this);
		install(PhenotypeCache.getPrototype());
		// Install a dummy value provider for the validate methods.
		install(new ValueProvider<Value<?>>(new BooleanValue(true)));

		for(Class<? extends DecoderModule<?>> module : enabledDecoders) {
			install(dseInjector.getInstance(module));
		}
	}
}
