/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.graph;

import java.util.Collection;
import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.opt4j.core.Genotype;
import org.opt4j.core.problem.Decoder;

import com.google.common.collect.Sets;
import com.google.inject.Injector;
import com.google.inject.Key;
import com.google.inject.Singleton;

/**
 * Finds the set of required {@link Phenotype}s that depend on a given set of {@link Genotype}s.
 * During construction, the set of demanded solution {@link Phenotype}s has to be given such that
 * the underlying dependency graph can be constructed.
 * For this graph, a visitor is constructed that remembers the dependencies of each
 * {@link Genotype} towards {@link Phenotype}s. The visitor traverses the decoding
 * {@link Injector} to capture the dependencies of the decoding phase.
 * This class is inspired by the InjectorGrapher of the Guice Grapher extension.
 * 
 * @author diewald
 */
@Singleton
public class DependentPhenotypeFinder extends BindingWalker<DependentPhenotypeVisitor> {

	/**
	 * Constructor.
	 * 
	 * @param phenotypes
	 *            Set of {@link Phenotype}s to be produced by the DSE.
	 * @param protoDecodingInjector
	 *            "Prototype" injector holding {@link Decoder}s enabled by features. Typically
	 *            constructed during problem module binding.
	 */
	public DependentPhenotypeFinder(Set<Class<? extends Phenotype>> phenotypes,
			Injector protoDecodingInjector) {
		super(new DependentPhenotypeVisitor());
		loadCache(phenotypes, protoDecodingInjector);
	}

	/**
	 * Triggers a walk over all bindings to capture the transitive dependencies of {@link Genotype}s
	 * towards {@link Phenotype}s
	 */
	private void loadCache(Set<Class<? extends Phenotype>> phenotypes, Injector injector) {
		Set<Key<?>> phenotypeKeys = Sets.newHashSet();
		phenotypes.forEach(p -> phenotypeKeys.add(Key.get(p)));
		iterateBindings(injector, phenotypeKeys);
	}

	/**
	 * For a given {@link Genotype}, this method finds all {@link Phenotype}s that depend on the
	 * given {@link Genotype} due to {@link Decoder}s. The result includes transitive dependencies.
	 * 
	 * @param modifiedGeno
	 *            Query {@link Genotype} (see above).
	 * @return Collection of dependent {@link Phenotype}s.
	 */
	public Collection<Class<? extends Phenotype>>
			getDependentPhenos(Class<? extends Genotype> modifiedGeno) {
		return keyVisitor.getDependentPhenos(modifiedGeno);
	}
}
