/*
 * Copyright (C) 2008 Google Inc.
 * Copyright (C) 2018 fortiss GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.graph;

import java.lang.reflect.Method;
import java.util.Collection;
import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.opt4j.core.Genotype;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;
import com.google.common.collect.Sets;
import com.google.inject.Binder;
import com.google.inject.Binding;
import com.google.inject.Key;
import com.google.inject.spi.Dependency;
import com.google.inject.spi.ElementSource;
import com.google.inject.spi.HasDependencies;

/**
 * Walks the dependency graph from a given set of {@link Genotype}s that must have a Guice binding
 * to obtain a mapping of {@link Genotype}s towards its relevant dependent {@link Phenotype}s. These
 * {@link Genotype}s have to be given at construction time.
 * The (transitive) dependencies of the {@link Genotype}s are returned by
 * {@link #getDependentPhenos(Class)}.
 * This class is a modified version of the TransitiveDependencyVisitor class of the Grapher
 * extension of google Guice.
 *
 * @author phopkins@gmail.com (Pete Hopkins)
 * @author diewald@fortiss.org (Alexander Diewald)
 */
public class DependentPhenotypeVisitor extends DependencyVisitorBase {

	/**
	 * Set of found types to which a dependency exists and which is compatible with any of the
	 * initially given type-set.
	 */
	private Multimap<Class<? extends Genotype>, Class<? extends Phenotype>> dependentPhenos =
			HashMultimap.create();

	/** Caches dependencies of {@link Phenotype}s to other {@link Phenotype}s. */
	private Multimap<Class<? extends Phenotype>, Class<? extends Phenotype>> internalPhenoDepCache =
			HashMultimap.create();

	/**
	 * Iterates over the dependency tree spanned by the {@link Binder} and remembers every
	 * encountered type binding if it is a sub-type of the initially given set of types.
	 */
	@SuppressWarnings("unchecked")
	@Override
	protected Collection<Key<?>> visitHasDependencies(HasDependencies hasDependencies,
			Binding<?> binding) {
		Class<? extends Phenotype> currProducedType = null;
		if(binding.getSource() instanceof ElementSource) {
			ElementSource source = (ElementSource)binding.getSource();
			Object declSource = source.getDeclaringSource();
			if(declSource instanceof Method) {
				Method declMethod = (Method)declSource;
				Class<?> prodClass = declMethod.getReturnType();
				if(Phenotype.class.isAssignableFrom(prodClass)) {
					currProducedType = (Class<? extends Phenotype>)prodClass;
				}
			}
		}

		Set<Key<?>> dependencies = Sets.newHashSet();
		for(Dependency<?> dependency : hasDependencies.getDependencies()) {
			Key<?> key = dependency.getKey();
			dependencies.add(key);
			// Just walk over other dependencies that do not produce Phenotypes.
			if(currProducedType != null) {
				addToDependencyCache(key, currProducedType);
			}
		}

		return dependencies;
	}

	/**
	 * Adds the type wrapped by the given {@link Key} to the set of found types if it is a subtype
	 * of any of the initially given type set.
	 */
	@SuppressWarnings("unchecked")
	private void addToDependencyCache(Key<?> key, Class<? extends Phenotype> currProducedType) {
		Class<?> depClass = key.getTypeLiteral().getRawType();
		if(Genotype.class.isAssignableFrom(depClass)) {
			dependentPhenos.put((Class<? extends Genotype>)depClass, currProducedType);
			// Also, add the transitive dependencies for the currently processed one.
			Collection<Class<? extends Phenotype>> prevPhenoDeps =
					internalPhenoDepCache.get(currProducedType);
			dependentPhenos.putAll((Class<? extends Genotype>)depClass, prevPhenoDeps);
		}
		if(Phenotype.class.isAssignableFrom(depClass)) {
			internalPhenoDepCache.put((Class<? extends Phenotype>)depClass, currProducedType);
			// Also, add the transitive dependencies for the currently processed one.
			Collection<Class<? extends Phenotype>> prevPhenoDeps =
					internalPhenoDepCache.get(currProducedType);
			internalPhenoDepCache.putAll((Class<? extends Phenotype>)depClass, prevPhenoDeps);
		}
	}

	/**
	 * Returns the set of types that are subtypes of the initial set of given types and to which a
	 * dependency exits.
	 */
	public Collection<Class<? extends Phenotype>>
			getDependentPhenos(Class<? extends Genotype> modifiedGeno) {
		return dependentPhenos.get(modifiedGeno);
	}
}
