/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem;

import static com.google.common.collect.Sets.newHashSet;
import static org.fortiss.af3.exploration.alg.service.ExplorationServiceManager.getService;
import static org.fortiss.af3.exploration.alg.util.ExplorationFeatureUtils.getEnabledApplicableFeatures;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import org.aopalliance.intercept.MethodInterceptor;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.GenericProcessOperatorModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.EncodingProviderModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.create.GuiceCreator;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.GuiceDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.evaluate.GuiceEvaluator;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.genotype.GenotypeProviderGenerator;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.graph.DecoderFinder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.graph.DependentPhenotypeFinder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.graph.GenotypeFinder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.graph.PhenotypeFinder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.PhenotypeProviderGenerator;
import org.fortiss.af3.exploration.alg.dse.module.CreatorModule;
import org.fortiss.af3.exploration.alg.dse.module.DecoderModule;
import org.fortiss.af3.exploration.alg.feature.ExplorationFeature;
import org.fortiss.af3.exploration.alg.feature.IExplorationFeature;
import org.fortiss.af3.exploration.alg.service.IExplorationFeatureService;
import org.fortiss.af3.exploration.moea.model.DseSpecification;
import org.opt4j.core.Genotype;
import org.opt4j.core.problem.Creator;
import org.opt4j.core.problem.Decoder;
import org.opt4j.core.problem.Evaluator;
import org.opt4j.core.problem.ProblemModule;

import com.google.inject.Injector;
import com.google.inject.Module;
import com.google.inject.TypeLiteral;
import com.google.inject.multibindings.Multibinder;

/**
 * Problem module that setups the main control entities constituting the DSE process exploration. It
 * defines the top-most {@link Creator}s, {@link Decoder}s, and {@link Evaluator}s, which consists
 * of sub-modules. Moreover, this class calculates the set of required decoders to be launched on
 * the basis of the activated {@link IExplorationFeature}s and the demanded solutions.
 * 
 * @author diewald
 */
public class DseProcessProblemModule extends ProblemModule {

	/** References the EA specification, also holding the exploration specification. */
	private DseSpecification dse;

	/** The set of solutions that shall be returned by the DSE. */
	private Set<Class<? extends Phenotype>> requestedSolutions;

	/**
	 * Reference to the DSE-wide {@link Injector} to allow access to the system model parameters.
	 */
	private Injector dseInjector;

	/** Constructor. */
	public DseProcessProblemModule(DseSpecification dse, Injector dseInjector,
			Set<Class<? extends Phenotype>> requestedSolutions) {
		this.dse = dse;
		this.dseInjector = dseInjector;
		this.requestedSolutions = requestedSolutions;
	}

	/**
	 * Configure {@link Genotype} creator (execution model dependent), problem decoder and goal
	 * evaluators.
	 */
	@Override
	protected void config() {
		GuiceEvaluator guiceEvaluator = dseInjector.getInstance(GuiceEvaluator.class);

		Collection<IExplorationFeature> enabledFeatures =
				getEnabledApplicableFeatures(dse, dseInjector, true);
		Collection<DecoderModule<?>> enabledDecoders = newHashSet();
		enabledFeatures.forEach(f -> enabledDecoders.addAll(f.getDecoders()));

		Injector protoDecodingInjector =
				createProtoDecodingInjector(enabledFeatures, enabledDecoders);
		Set<Class<? extends Genotype>> requiredGenotypes =
				findRequiredGenotypes(protoDecodingInjector);

		// Creators & Genotype Provider Module
		GuiceCreator guiceCreator = createGuiceCreator(enabledFeatures, requiredGenotypes);
		bind(GenotypeProviderGenerator.class)
				.toInstance(new GenotypeProviderGenerator(requiredGenotypes));

		// Decoders, Phenotype dependencies & Phenotype Provider Module
		GuiceDecoder guiceDecoder = createGuiceDecoder(protoDecodingInjector);
		DependentPhenotypeFinder depPhenoFinder =
				new DependentPhenotypeFinder(requestedSolutions, protoDecodingInjector);
		bind(DependentPhenotypeFinder.class).toInstance(depPhenoFinder);
		Set<Class<? extends Phenotype>> constructPhenotype =
				findConstructedPhenotypes(protoDecodingInjector);
		constructPhenotype.addAll(requestedSolutions);
		bind(PhenotypeProviderGenerator.class)
				.toInstance(new PhenotypeProviderGenerator(constructPhenotype));

		bindProcessProblem(guiceCreator, guiceDecoder, guiceEvaluator);
	}

	/**
	 * Creates a decoding injector that contains modules providing all {@link Genotype}s from the
	 * given set of enabled features and the set of given {@link Decoder}s.
	 * 
	 * @param enabledFeatures
	 *            {@link ExplorationFeature}s enabled for this execution of the D3SE.
	 * @param enabledDecoders
	 *            Set of decoders to install.
	 * @return {@link Injector} hosting the discussed modules.
	 */
	private Injector createProtoDecodingInjector(Collection<IExplorationFeature> enabledFeatures,
			Collection<DecoderModule<?>> enabledDecoders) {
		// We must create and install the Genotype Prototype Providers here to allow Guice analysing
		// the dependencies correctly (all available Genotypes are used therefore).
		Set<Class<? extends Genotype>> availableGenotypes =
				getService(IExplorationFeatureService.class)
						.getCreatableGenotypesOf(enabledFeatures);
		EncodingProviderModule<Genotype> protoGenoProviderModule =
				new GenotypeProviderGenerator(availableGenotypes).getEncodingProviderModule();
		GuiceDecoder dependencyAnalysisDecoder = createDependencyAnalysisDecoder(enabledDecoders);
		Injector decodingInjector =
				dseInjector.createChildInjector(protoGenoProviderModule, dependencyAnalysisDecoder);
		return decodingInjector;
	}

	/**
	 * Creates a {@link GuiceCreator}, prepares all {@link CreatorModule}s, and defines the
	 * {@link MethodInterceptor} for them.
	 * 
	 * @param enabledFeatures
	 *            Set of {@link IExplorationFeature}s enabled for this DSE execution.
	 * @param requiredGenotypes
	 *            Set of {@link Genotype}s required for the current configuration of the DSE:
	 * @return The created/configured {@link CreatorModule}s.
	 */
	@SuppressWarnings("unchecked")
	private GuiceCreator createGuiceCreator(Collection<IExplorationFeature> enabledFeatures,
			Set<Class<? extends Genotype>> requiredGenotypes) {
		Collection<Class<? extends CreatorModule<?>>> requiredCreators = new HashSet<>();
		for(IExplorationFeature feature : enabledFeatures) {
			for(Class<? extends Genotype> reqGenotype : requiredGenotypes) {
				CreatorModule<?> reqCreator = feature.getCreatorOf(reqGenotype);
				if(reqCreator != null) {
					// The actually used CreatorModule must be created by Guice to enable method
					// interception. This is done in the GuiceCreator module.
					requiredCreators.add((Class<? extends CreatorModule<?>>)reqCreator.getClass());
				}
			}
		}
		GuiceCreator guiceCreator = dseInjector.getInstance(GuiceCreator.class);
		guiceCreator.setEnabledCreators(requiredCreators);
		return guiceCreator;
	}

	/**
	 * Creates a {@link GuiceDecoder}, prepares all {@link DecoderModule}s, and defines the
	 * {@link MethodInterceptor} for them, by creating these {@link Module}s from Guice. The
	 * intercept mechanism would be non-operational if {@link Module}s are constructed by the "new"
	 * operator.
	 * 
	 * @param decodingInjector
	 *            "Prototype" {@link Injector} hosting all decoders enabled by
	 *            {@link ExplorationFeature}s.
	 * @return The created/configured {@link DecoderModule}s.
	 */
	private GuiceDecoder createGuiceDecoder(Injector decodingInjector) {
		Set<Class<? extends DecoderModule<?>>> requiredDecoders =
				findRequiredDecoders(decodingInjector);
		GuiceDecoder guiceDecoder = dseInjector.getInstance(GuiceDecoder.class);
		guiceDecoder.setEnabledDecoders(requiredDecoders);
		return guiceDecoder;
	}

	/**
	 * Creates a {@link GuiceDecoder}, prepares all {@link DecoderModule}s, and defines the
	 * {@link MethodInterceptor} for plugin-type {@link DecoderModule}s.
	 * 
	 * @param enabledDecoders
	 *            Set of {@link DecoderModule}s configure for this DSE execution.
	 * @return The created/configured {@link GuiceDecoder}.
	 */
	@SuppressWarnings("unchecked")
	private GuiceDecoder
			createDependencyAnalysisDecoder(Collection<DecoderModule<?>> enabledDecoders) {
		Collection<Class<? extends DecoderModule<?>>> enabledDecoderTypes = new HashSet<>();
		enabledDecoders.forEach(
				dec -> enabledDecoderTypes.add((Class<? extends DecoderModule<?>>)dec.getClass()));
		GuiceDecoder guiceDecoder = dseInjector.getInstance(GuiceDecoder.class);
		guiceDecoder.setEnabledDecoders(enabledDecoderTypes);
		return guiceDecoder;
	}

	/**
	 * Returns the set of {@link Genotype}s that are required as inputs to decoders to obtain
	 * the requested {@link Phenotype}.
	 */
	private Set<Class<? extends Genotype>> findRequiredGenotypes(Injector decodingInjector) {
		GenotypeFinder genotypeFinder = new GenotypeFinder();
		return genotypeFinder.getGenotypesFor(requestedSolutions, decodingInjector);
	}

	/**
	 * Returns the set of {@link DecoderModule}s that is required to decode the set of
	 * {@link #requestedSolutions}. These are the {@link DecoderModule}s which are used in the DSE
	 * run.
	 */
	private Set<Class<? extends DecoderModule<?>>> findRequiredDecoders(Injector decodingInjector) {
		DecoderFinder decoderFinder = new DecoderFinder();
		return decoderFinder.getDecodersFor(requestedSolutions, decodingInjector);
	}

	/**
	 * Returns the set of {@link DecoderModule}s that is required to decode the set of
	 * {@link #requestedSolutions}. These are the {@link DecoderModule}s which are used in the DSE
	 * run.
	 */
	private Set<Class<? extends Phenotype>> findConstructedPhenotypes(Injector decodingInjector) {
		PhenotypeFinder phenotypeFinder = new PhenotypeFinder();
		return phenotypeFinder.getProducedPhenosFor(requestedSolutions, decodingInjector);
	}

	/**
	 * Binds the topmost {@link Creator}, {@link Decoder}, and {@link Evaluator} that orchestrate
	 * the dependency-driven execution of their submodules.
	 */
	@SuppressWarnings("unchecked")
	private void bindProcessProblem(Creator<?> creator, Decoder<?, ?> decoder,
			Evaluator<?> evaluator) {
		// This is inspired from ProblemModule.bindProblem(), which does not allow to pass
		// parameters to the constructor of the creators/decoders. Hence, a pre-constructed
		// creator/decoder is bound here.

		// Note that in contrast to ProblemModule.bindProblem(), the creator and the decoder
		// are not bound in the singleton scope, since these classes need to be parameterized
		// with the current input model.
		binder().bind(new TypeLiteral<Creator<Genotype>>() { // empty
		}).toInstance((Creator<Genotype>)creator);

		binder().bind(new TypeLiteral<Decoder<Genotype, Object>>() { // empty
		}).toInstance((Decoder<Genotype, Object>)decoder);

		Multibinder<Evaluator<Object>> multibinder =
				Multibinder.newSetBinder(binder(), new TypeLiteral<Evaluator<Object>>() { //
				});
		multibinder.addBinding().toInstance((Evaluator<Object>)evaluator);

		install(new GenericProcessOperatorModule());
	}
}
