/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.modeltransformation.base;

import static org.fortiss.tooling.common.util.LambdaUtils.filterList;
import static org.fortiss.tooling.common.util.LambdaUtils.filterSet;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getChildrenWithType;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickInstanceOf;

import java.util.Collection;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.ITransformationStrategy;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.element.IElementTransformationModule;
import org.fortiss.af3.exploration.alg.exception.TransformationModuleException;
import org.fortiss.af3.exploration.alg.guava.SubClassToInstanceMap;
import org.fortiss.af3.exploration.alg.service.IExplorationTransformationService.TransformationContext;
import org.fortiss.af3.exploration.alg.service.module.ExplorationTransformationInputs;
import org.fortiss.af3.project.model.FileProject;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;

/**
 * 
 * @author diewald
 */
public abstract class ElementTransformationStrategyBase<T extends EObject, O extends EObject>
		extends TransformationStrategyBase<T> {

	/** Contains the order to execute the {@link IElementTransformationModule}s. */
	protected List<IElementTransformationModule<EObject, O>> execOrder = new LinkedList<>();

	/** {@inheritDoc} */
	@Override
	public <N extends EObject, S extends EObject> T transform(TransformationState state,
			TransformationContext context, ExplorationTransformationInputs inputs,
			SubClassToInstanceMap<N> transformedModels, EObject outputModel)
			throws TransformationModuleException {
		// TODO: Consider using generics to define the allowed sub-module types.

		// TODO: Throw an exception if the output model is not the same type as the input model.
		@SuppressWarnings("unchecked") T castedTransformedModel = (T)outputModel;
		this.transformedModel = castedTransformedModel;

		if(execOrder == null || execOrder.isEmpty()) {
			calculateModuleExecutionOrder();
		}

		for(IElementTransformationModule<? extends EObject, O> subModule : execOrder) {
			if(subModule.getActivationStage() == state ||
					subModule instanceof ElementTransformationStrategyBase) {
				if(subModule.isOptional() &&
						!canExecuteModule(inputs, transformedModels, subModule)) {
					break;
				}

				if(subModule instanceof ITransformationStrategy) {
					((ITransformationStrategy<?>)subModule).setModelContext(modelContext);
					((ITransformationStrategy<?>)subModule).setLaunchedAsSubModule(true);
				}

				// TODO: directly specify the output.
				@SuppressWarnings("unchecked") O tempCastedOutputModel =
						(O)transformedModels.get(subModule.getElementContainerType());

				subModule.applyAssociationUpdates(this);
				T transformedModel = transformElements(subModule, inputs, transformedModels,
						tempCastedOutputModel);
				if(transformedModel == null) {
					throw new TransformationModuleException(subModule,
							"The submodule did not return the transformed model. It is required by the framework.");
				}

				// Impl classes in EMF _always_ have one interface that specifies the methods etc.
				@SuppressWarnings("unchecked") Class<N> transformedModelType =
						(Class<N>)transformedModel.getClass().getInterfaces()[0];
				@SuppressWarnings("unchecked") N tempCastedTransformedModel = (N)transformedModel;
				transformedModels.put(transformedModelType, tempCastedTransformedModel);
				if(getOutputType().isAssignableFrom(transformedModel.getClass())) {
					this.transformedModel = transformedModel;
				}
				applyAssociationUpdates(subModule);
			}
		}

		executeSubModules(
				filterSet(getSubModules(), e -> !(e instanceof IElementTransformationModule<?, ?>)),
				context, inputs, transformedModels, state);

		@SuppressWarnings("unchecked") Class<N> tempTransformedModelType =
				(Class<N>)transformedModel.getClass().getInterfaces()[0];
		@SuppressWarnings("unchecked") N tempCastedTransformedModel = (N)transformedModel;
		transformedModels.put(tempTransformedModelType, tempCastedTransformedModel);
		if(getOutputType().isAssignableFrom(transformedModel.getClass())) {
			@SuppressWarnings("unchecked") T castedOutputModel = (T)outputModel;
			this.transformedModel = castedOutputModel;
		}

		// Suppressed warnings: Generics are used here to ensure compatibility between the type
		// (class) and the instance.
		@SuppressWarnings("unchecked") Class<? extends N> transformedModelType =
				(Class<N>)getOutputType();
		@SuppressWarnings("unchecked") N castedOutputModel = (N)transformedModel;
		transformedModels.put(transformedModelType, castedOutputModel);
		@SuppressWarnings("unchecked") T castedReturnOutputModel = (T)outputModel;
		return castedReturnOutputModel;
	}

	/**
	 * Calculates the correct order of execution of element-wise transformation sub-modules.
	 * It does so by building a graph of the require inputs and output types of the element-wise
	 * transformation sub modules.
	 */
	@SuppressWarnings("unchecked")
	protected void calculateModuleExecutionOrder() {
		execOrder.addAll((Collection<? extends IElementTransformationModule<EObject, O>>)filterList(
				getSubModules(), e -> e instanceof IElementTransformationModule));
		execOrder.sort(new ModuleExecOrderComp());
	}

	/**
	 * Comparator that sorts a set of {@link IElementTransformationModule}s first by the order of
	 * the
	 * {@link org.fortiss.af3.exploration.alg.dse.modeltransformation.ITransformationModule.TransformationState}
	 * which activates them, and then by the order of their required
	 * inputs.
	 * This implies that a {@link IElementTransformationModule} may define that all other modules of
	 * the same
	 * {@link org.fortiss.af3.exploration.alg.dse.modeltransformation.ITransformationModule.TransformationState}
	 * which output elements of a given type to be executed before the current one.
	 */
	protected class ModuleExecOrderComp
			implements Comparator<IElementTransformationModule<EObject, O>> {
		/** {@inheritDoc} */
		@Override
		public int compare(IElementTransformationModule<EObject, O> arg0,
				IElementTransformationModule<EObject, O> arg1) {
			int result = arg0.getActivationStage().compareTo(arg1.getActivationStage());

			if(result == 0) {
				Collection<Class<?>> inputs0 = arg0.getRequiredTypes();
				Collection<Class<?>> inputs1 = arg1.getRequiredTypes();
				if(inputs1.contains(arg0.getOutputType())) {
					return -1;
				} else if(inputs0.contains(arg1.getOutputType())) {
					return 1;
				}
			}

			return result;
		}
	}

	/**
	 * Triggers the registered {@link IElementTransformationModule}s for each {@link EObject} of the
	 * module's demanded type found in the input model.
	 * 
	 * @param subModule
	 *            {@link IElementTransformationModule} implementing the element-wise transformation.
	 * @param inputs
	 *            Original input models containing information for the transformation.
	 * @param transformedModels
	 *            Optional map of transformed/generated models in the transformation process.
	 * @param outputModel
	 *            The effective transformation model to operate on.
	 * @return The transformed {@code outputModel}.
	 * @throws TransformationModuleException
	 *             If any of the submodules fails.
	 */
	protected <N extends EObject, S extends EObject> T transformElements(
			IElementTransformationModule<? extends EObject, O> subModule,
			ExplorationTransformationInputs inputs, Map<Class<? extends N>, N> transformedModels,
			O outputModel) throws TransformationModuleException {
		Collection<? extends EObject> typedElements =
				getChildrenWithType(outputModel, subModule.getElementClass());
		for(EObject element : typedElements) {
			subModule.transformElement(element, inputs, transformedModels, outputModel);
		}

		@SuppressWarnings("unchecked") T castedOutputModel = (T)outputModel;
		return castedOutputModel;
	}

	/** {@inheritDoc} */
	@Override
	public void addModelToProject(FileProject fp, ITopLevelElement modelContext)
			throws TransformationModuleException {
		this.modelContext = modelContext;
		for(ITransformationStrategy<?> module : pickInstanceOf(ITransformationStrategy.class,
				subModules)) {
			module.addModelToProject(fp, modelContext);
		}
	}
}
