/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.service;

import java.util.Collection;
import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.module.EvaluatorModule;
import org.fortiss.af3.exploration.alg.dse.module.IEvaluatorModule;
import org.fortiss.af3.exploration.alg.feature.IExplorationFeature;
import org.opt4j.core.Genotype;

/**
 * Service for adding {@link IExplorationFeature}s to the DSE. They define the concrete problems
 * that shall be solved by the DSE process. {@link IExplorationFeature}s registered with this
 * service are known to the DSE and can be selected by users to define a DSE process.
 * 
 * Each feature consists of so-called Exploration Modules that can be Variables, Decodings of these
 * Variables (aka. Problem descriptions), or evaluators that rate {@link Phenotype}s to derive
 * metrics for the quality of a solution.
 * 
 * @author diewald
 */
public interface IExplorationFeatureService extends IExplorationService {

	/**
	 * Registers the given {@link IExplorationFeature} with this service to make it selectable by
	 * users of the DSE.
	 * 
	 * @param feature
	 *            {@link IExplorationFeature} to register with the DSE.
	 */
	void registerExplorationFeature(IExplorationFeature feature);

	/** Returns the registered {@link IExplorationFeature}s. */
	Collection<IExplorationFeature> getExplorationFeatures();

	/**
	 * Returns the set of {link {@link Genotype}s that can be produced by the given set of features.
	 */
	Set<Class<? extends Genotype>>
			getCreatableGenotypesOf(Collection<IExplorationFeature> features);

	/**
	 * Registers the given {@link IEvaluatorModule} with the overall set of exploration modules and
	 * checks for double registrations.
	 * 
	 * @param evaluatorModule
	 *            {@link EvaluatorModule} to register.
	 */
	void registerEvaluatorModule(IEvaluatorModule<?, ?, ?> evaluatorModule);

	/** Returns all registered {@link IEvaluatorModule}s. */
	Set<IEvaluatorModule<?, ?, ?>> getEvaluatorModules();

	/**
	 * Returns the {@link IExplorationFeature}s that have the given types from the set of registered
	 * {@link IExplorationFeature}s.
	 */
	Collection<IExplorationFeature>
			getExplorationFeatures(Collection<Class<? extends IExplorationFeature>> features);
}
