/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.service.internal;

import static java.util.Collections.emptyList;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import org.fortiss.af3.exploration.alg.exception.ExplorationException;
import org.fortiss.af3.exploration.alg.service.IExpSpecTransformationService;
import org.fortiss.af3.exploration.alg.service.module.IExpSpecElementTransformator;
import org.fortiss.af3.exploration.alg.service.module.IExpSpecTransformationModule;
import org.fortiss.af3.exploration.alg.service.module.base.ExpSpecTransformationModuleBase;
import org.fortiss.af3.exploration.model.ExplorationSpecification;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.tooling.kernel.introspection.IIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectionItem;

import com.google.inject.Injector;

/**
 * Provides the {@link IExpSpecTransformationService}.
 * 
 * @author diewald
 */
public class ExpSpecTransformationService extends ExpSpecTransformationModuleBase
		implements IExpSpecTransformationService {

	/** Holds the instance of the evaluation service. */
	private volatile static ExpSpecTransformationService INSTANCE;

	/** The instance (singleton) of the ExplorationTransformationService. */
	public static synchronized ExpSpecTransformationService getInstance() {
		if(INSTANCE == null) {
			INSTANCE = new ExpSpecTransformationService();
		}
		return INSTANCE;
	}

	/** Constructor. */
	private ExpSpecTransformationService() {
		// Prevent direct instantiation.
	}

	/** {@inheritDoc} */
	@Override
	public Map<ExplorationTarget<?>, ExplorationTarget<?>> transform(Injector mainInjector,
			Map<ExplorationTarget<?>, ExplorationTarget<?>> inputExpTargets)
			throws ExplorationException {
		// From the service "downwards" we must ensure the existence of a "real" map, not views or
		// sth. similar, so reaquire a map here.
		return super.transform(mainInjector, new HashMap<>(inputExpTargets));
	}

	/** {@inheritDoc} */
	@Override
	public void initializeService() {
		// Nothing to do.
	}

	/** {@inheritDoc} */
	@Override
	public void startService() {
		// Nothing to do.
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionDescription() {
		return "The " + getClass().getSimpleName() + " service allows to define transformation " +
				"transformation rules for " + ExplorationSpecification.class.getSimpleName() +
				"that are executed before the D3SE is actually launched. These rules are encoded " +
				"in " + IExpSpecTransformationModule.class.getSimpleName() + "s that may be " +
				"composed at arbitrary depth levels. These modules are allowed to use " +
				IExpSpecElementTransformator.class.getSimpleName() + "s that process one element " +
				"per execution provided by the parent " +
				IExpSpecTransformationModule.class.getSimpleName() + ".";
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return "Exploration Specification Transformation Service";
	}

	/** {@inheritDoc} */
	@Override
	public boolean showInIntrospectionNavigation() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<IIntrospectionItem> getIntrospectionItems() {
		return emptyList();
	}

	/** {@inheritDoc} */
	@Override
	public IIntrospectionDetailsItem getDetailsItem() {
		return null;
	}
}
