/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.service.module.base;

import static org.fortiss.tooling.kernel.utils.EcoreUtils.getChildrenWithType;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickInstanceOf;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Map;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.exploration.alg.exception.ExplorationException;
import org.fortiss.af3.exploration.alg.service.module.IExpSpecElementTransformator;
import org.fortiss.af3.exploration.alg.service.module.IExpSpecTransformationModule;
import org.fortiss.af3.exploration.model.ExplorationTarget;

import com.google.inject.Injector;

/**
 * Base class for {@link IExpSpecTransformationModule}s whose children are
 * {@link IExpSpecElementTransformator}s. This class manages extracting and providing the requested
 * input elements to the children.
 * 
 * @author diewald
 */
public abstract class ExpSpecElementTransformationManager
		implements IExpSpecTransformationModule<IExpSpecElementTransformator<?, ?>> {

	/** Set of {@link IExpSpecTransformationModule}s registered in this module */
	@SuppressWarnings("rawtypes")
	protected Collection<IExpSpecElementTransformator> modules = new ArrayList<>();

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public Map<ExplorationTarget<?>, ExplorationTarget<?>> transform(Injector mainInjector,
			Map<ExplorationTarget<?>, ExplorationTarget<?>> inputExpTargets)
			throws ExplorationException {
		Map<ExplorationTarget<?>, ExplorationTarget<?>> transformedMap = inputExpTargets;
		for(IExpSpecElementTransformator<EObject, ?> module : modules) {
			Collection<EObject> inputElements = new ArrayList<>();
			Class<EObject> inputType = module.getInputType();
			pickInstanceOf(inputType, new ArrayList<>(inputExpTargets.keySet())).stream()
					.forEach(inputElements::add);
			inputExpTargets.keySet().stream()
					.flatMap(t -> getChildrenWithType(t, inputType).stream())
					.forEach(inputElements::add);
			for(EObject inputElement : inputElements) {
				Object out = module.transform(mainInjector, inputElement);
				// TODO(AD): consider making this logic more generic, maybe at the cost of generics
				// of the map.
				if(inputElement instanceof ExplorationTarget && out instanceof ExplorationTarget) {
					transformedMap.put((ExplorationTarget<?>)inputElement,
							(ExplorationTarget<?>)out);
				}
			}
		}
		return transformedMap;
	}

	/** {@inheritDoc} */
	@Override
	public void registerModule(IExpSpecElementTransformator<?, ?> module) {
		modules.add(module);
	}
}
