/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.util;

import static org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceConnectionAdapter.ConnectionType.BIDIRECTIONAL;
import static org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceConnectionAdapter.ConnectionType.INCOMING;
import static org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceConnectionAdapter.ConnectionType.OUTGOING;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getParentsWithType;
import static org.fortiss.tooling.kernel.utils.KernelModelElementUtils.getParentElement;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Optional;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.component.model.Channel;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.component.model.ComponentArchitecture;
import org.fortiss.af3.component.model.InputPort;
import org.fortiss.af3.component.model.OutputPort;
import org.fortiss.af3.component.model.Port;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.safetyfunctionarch.SafetyFunctionArchEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IModelElementAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceConnectionAdapter.ConnectionType;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.IMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.IMappingEntry;
import org.fortiss.af3.exploration.alg.exception.InvalidModelException;
import org.fortiss.af3.exploration.alg.exception.InvalidPlatformModelException;
import org.fortiss.af3.exploration.alg.exception.InvalidTimingModelException;
import org.fortiss.af3.exploration.alg.guava.TypedBiMap;
import org.fortiss.af3.platform.model.IPlatformPort;
import org.fortiss.af3.platform.model.PlatformConnectorUnit;
import org.fortiss.af3.safety.model.SafetyArchitecture;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IConnection;
import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.utils.AnnotationUtils;
import org.fortiss.tooling.kernel.model.IIdLabeled;
import org.jgrapht.graph.DefaultDirectedGraph;
import org.jgrapht.graph.DefaultEdge;

/**
 * Utility methods which help to abstract AF3 specifica for the DSE.
 * 
 * @author diewald
 */
public class AF3Utils {

	/** Constructor. Prevent initialization. */
	private AF3Utils() {
	}

	/**
	 * Translates the safety architecture encoded in a {@link SafetyFunctionArchEntry} to a
	 * {@link SafetyArchitecture} enum representation.
	 * 
	 * @param sfaEntry
	 *            {@link SafetyFunctionArchEntry} to be translated.
	 * @return The encoded {@link SafetyArchitecture}, or {@code null} if no matching enum is found.
	 */
	public static SafetyArchitecture getSFArchFromEntry(SafetyFunctionArchEntry sfaEntry) {
		int channelNum = sfaEntry.getNumChannels();
		int triggerNum = sfaEntry.getNumTriggerSignal();
		boolean diagnosisPresent = sfaEntry.isDiagnosicUnitPresent();
		boolean diagnosisConnected = sfaEntry.isDiagnosicUnitConnected();

		if(diagnosisConnected) {
			if(channelNum == 1 && triggerNum == 1) {
				return SafetyArchitecture.ARCH_1OO1_D;
			} else if(channelNum == 2 && triggerNum == 1) {
				return SafetyArchitecture.ARCH_1OO2_D;
			}
		} else {
			if(channelNum == 1 && triggerNum == 1) {
				return diagnosisPresent ? SafetyArchitecture.ARCH_1OO1_DU
						: SafetyArchitecture.ARCH_1OO1;
			} else if(channelNum == 2 && triggerNum == 1) {
				return diagnosisPresent ? SafetyArchitecture.ARCH_1OO2_DU
						: SafetyArchitecture.ARCH_1OO2;
			} else if(channelNum == 3 && triggerNum == 1) {
				return diagnosisPresent ? SafetyArchitecture.ARCH_1OO3_DU
						: SafetyArchitecture.ARCH_1OO3;
			} else if(channelNum == 2 && triggerNum == 2) {
				return diagnosisPresent ? SafetyArchitecture.ARCH_2OO2_DU
						: SafetyArchitecture.ARCH_2OO2;
			} else if(channelNum == 3 && triggerNum == 2) {
				return diagnosisPresent ? SafetyArchitecture.ARCH_2OO3_DU
						: SafetyArchitecture.ARCH_2OO3;
			} else if(channelNum == 4 && triggerNum == 2) {
				return SafetyArchitecture.ARCH_2OO4;
			}
		}

		return null;
	}

	/**
	 * Returns the number of independent safety channels encoded in a {@link SafetyArchitecture}.
	 * 
	 * @param sfArch
	 *            {@link SafetyArchitecture} from which the channel number shall be extracted.
	 * @return Number of channels, or {@code null} if no {@link SafetyArchitecture} is given or the
	 *         architecture is unknown.
	 */
	public static Integer getChannelNumbersSafetyArch(SafetyArchitecture sfArch) {
		if(sfArch == null) {
			return null;
		}

		switch(sfArch) {
			case ARCH_1OO1:
			case ARCH_1OO1_D:
			case ARCH_1OO1_DU:
				return 1;
			case ARCH_1OO2:
			case ARCH_1OO2_D:
			case ARCH_1OO2_DU:
			case ARCH_2OO2:
			case ARCH_2OO2_DU:
				return 2;
			case ARCH_1OO3:
			case ARCH_1OO3_DU:
			case ARCH_2OO3:
			case ARCH_2OO3_DU:
				return 3;
			case ARCH_2OO4:
				return 4;
			default:
				return null;
		}
	}

	/**
	 * Returns the number signals from independent safety channels that are required to trigger the
	 * actual safety function. This information is extracted form the given
	 * {@link SafetyArchitecture}.
	 * 
	 * @param sfArch
	 *            {@link SafetyArchitecture} from which the number of triggers shall be extracted.
	 * @return Number of triggers, or {@code null} if no {@link SafetyArchitecture} is given or the
	 *         architecture is unknown.
	 */
	public static Integer getTriggerNumbersSafetyArch(SafetyArchitecture sfArch) {
		if(sfArch == null) {
			return null;
		}

		switch(sfArch) {
			case ARCH_1OO1:
			case ARCH_1OO1_D:
			case ARCH_1OO1_DU:
			case ARCH_1OO2:
			case ARCH_1OO2_D:
			case ARCH_1OO2_DU:
			case ARCH_1OO3:
			case ARCH_1OO3_DU:
				return 1;
			case ARCH_2OO2:
			case ARCH_2OO2_DU:
			case ARCH_2OO3:
			case ARCH_2OO3_DU:
			case ARCH_2OO4:
				return 2;
			default:
				return null;
		}
	}

	/**
	 * Returns whether the given {@link SafetyArchitecture} encodes the presence of one or more
	 * diagnostic units.
	 * 
	 * @param sfArch
	 *            {@link SafetyArchitecture} encoding the presence of diagnostic unit (s).
	 * @return Presence of diagnostic unit(s), or {@code null} if no {@link SafetyArchitecture} is
	 *         given or the architecture is unknown.
	 */
	public static Optional<Boolean> isDiagPresentSafetyArch(SafetyArchitecture sfArch) {
		switch(sfArch) {
			case ARCH_1OO1_D:
			case ARCH_1OO1_DU:
			case ARCH_1OO2_D:
			case ARCH_2OO2_DU:
			case ARCH_1OO2_DU:
			case ARCH_1OO3_DU:
			case ARCH_2OO3_DU:
				return Optional.of(true);
			case ARCH_1OO1:
			case ARCH_1OO2:
			case ARCH_2OO2:
			case ARCH_1OO3:
			case ARCH_2OO3:
			case ARCH_2OO4:
				return Optional.of(false);
			default:
				return Optional.empty();
		}
	}

	/**
	 * Returns whether the given {@link SafetyArchitecture} encodes the presence of one or more
	 * diagnostic units that directly influence whether the safety function is triggered.
	 * 
	 * @param sfArch
	 *            {@link SafetyArchitecture} encoding the presence of diagnostic unit (s).
	 * @return Presence of diagnostic unit(s) influencing the safety function's output, or
	 *         {@code null} if no {@link SafetyArchitecture} is given or the architecture is
	 *         unknown.
	 */
	public static Optional<Boolean> isDiagConnectedSafetyArch(SafetyArchitecture sfArch) {
		switch(sfArch) {
			case ARCH_1OO1_D:
			case ARCH_1OO2_D:
				return Optional.of(true);
			case ARCH_1OO1:
			case ARCH_1OO1_DU:
			case ARCH_1OO2:
			case ARCH_1OO2_DU:
			case ARCH_2OO2:
			case ARCH_2OO2_DU:
			case ARCH_1OO3:
			case ARCH_1OO3_DU:
			case ARCH_2OO3:
			case ARCH_2OO3_DU:
			case ARCH_2OO4:
				return Optional.of(false);
			default:
				return Optional.empty();
		}
	}

	/**
	 * Determines the {@link ConnectionType} of a connection by examining the class of the given
	 * connector that is the target of the current connection.
	 * 
	 * @param sourceConnector
	 *            "sending" connector.
	 * @param targetConnector
	 *            Target {@link IConnector} of the {@link IConnection} whose direction shall be
	 *            determined.
	 * @param systemModelAdapter
	 *            Abstraction of the input models.
	 * @return {@link ConnectionType} specifying the direction of the {@link IConnection}.
	 * @throws InvalidPlatformModelException
	 *             if an invalid combination of of platform ports/interfaces has been detected
	 *             (e.g., Input <-> Input).
	 */
	// TODO: update doc.
	public static ConnectionType getConnectionTypeDirection(IConnector sourceConnector,
			IConnector targetConnector, SystemModelAdapter<?, ?, ?, ?, ?, ?, ?> systemModelAdapter)
			throws InvalidPlatformModelException {
		// By the method signature, we know the instance type is a IConnector class.
		@SuppressWarnings("unchecked") ConnectionType connectionTypeSource =
				systemModelAdapter.getConnectionType(
						(Class<? extends IConnector>)sourceConnector.eClass().getInstanceClass());
		@SuppressWarnings("unchecked") ConnectionType connectionTypeTarget =
				systemModelAdapter.getConnectionType(
						(Class<? extends IConnector>)targetConnector.eClass().getInstanceClass());
		// assert (connectionTypeSource != null && connectionTypeTarget != null) :
		// "Could not identify the type of one of the connection targets.";
		if((connectionTypeSource == INCOMING && connectionTypeTarget == INCOMING) ||
				(connectionTypeSource == OUTGOING && connectionTypeTarget == OUTGOING)) {
			throw new InvalidPlatformModelException((PlatformConnectorUnit)sourceConnector,
					"A platform port or interface (see above) for incoming or outgoing traffic is connected to an port or interface of the same type. This is not allowed.");
		}

		if(connectionTypeSource == OUTGOING && connectionTypeTarget == INCOMING) {
			return OUTGOING;
		} else if(connectionTypeSource == INCOMING && connectionTypeTarget == OUTGOING) {
			return INCOMING;
		} else if((connectionTypeSource == BIDIRECTIONAL || connectionTypeSource == null) &&
				(connectionTypeTarget != null && connectionTypeTarget != BIDIRECTIONAL)) {
			// Revert the direction for the target port, since we are examining the connection from
			// its source.
			return connectionTypeTarget == INCOMING ? OUTGOING : INCOMING;
		} else if((connectionTypeTarget == BIDIRECTIONAL || connectionTypeTarget == null) &&
				(connectionTypeSource != null && connectionTypeSource != BIDIRECTIONAL)) {
			return connectionTypeSource;
		}
		return BIDIRECTIONAL;
	}

	/**
	 * Returns the given {@link IAnnotatedSpecification} type of the given {@link IModelElement} and
	 * throws an exception if the annotation cannot be found.
	 * 
	 * @param element
	 *            {@link IModelElement} to fetch the {@link IAnnotatedSpecification} from.
	 * @param annotationClass
	 *            {@link IAnnotatedSpecification}-Type.
	 * @return The {@link IAnnotatedSpecification} of the given {@link IModelElement}.
	 * @throws InvalidModelException
	 *             if no {@link IAnnotatedSpecification} of the given type exists for the given
	 *             {@link IModelElement}.
	 */
	public static <T extends IAnnotatedSpecification> T getAnnotationChecked(IModelElement element,
			Class<T> annotationClass) throws InvalidModelException {
		T annotation = AnnotationUtils.getAnnotation(element, annotationClass);
		if(annotation == null) {
			throw new InvalidModelException(element, annotationClass);
		}
		return annotation;
	}

	/**
	 * Returns the type of communication the given port can handle: incoming, outgoing, or
	 * bidirectional traffic.
	 * 
	 * @param sourceConnector
	 *            {@link IConnector} to be examined.
	 * @param systemModelAdapter
	 *            Abstraction of the input models.
	 * @return The port type (default = bidirectional).
	 */
	public static ConnectionType getPortDirection(IPlatformPort sourceConnector,
			SystemModelAdapter<?, ?, ?, ?, ?, ?, ?> systemModelAdapter) {
		// By the method signature, we know the instance type is a IConnector class.
		@SuppressWarnings("unchecked") ConnectionType connectionTypeSource =
				systemModelAdapter.getConnectionType(
						(Class<? extends IConnector>)sourceConnector.eClass().getInstanceClass());
		return connectionTypeSource != null ? connectionTypeSource : BIDIRECTIONAL;
	}

	/**
	 * Predicate whether the subject {@link PlatformConnectorUnit} is a connector of the given
	 * {@link IModelElementAdapter}.
	 */
	public static boolean isConnectorOf(PlatformConnectorUnit pConn,
			IModelElementAdapter<?> element) {
		return pConn.getOwner() == element.getObject();
	}

	/**
	 * Predicate whether the subject {@link PlatformConnectorUnit}'s containing element is contained
	 * in the given {@link IModelElementAdapter} and also contained in the given
	 * {@link DefaultDirectedGraph}.
	 */
	public static boolean isContainedInResource(PlatformConnectorUnit pConn,
			IModelElementAdapter<?> element) {
		return pConn.getOwner() == ((IHierarchicElement)element.getObject()).getContainer();
	}

	/**
	 * Predicate whether the subject {@link PlatformConnectorUnit} is the predecessor
	 * {@link PlatformConnectorUnit} for the device-internal communication of the given
	 * {@link IModelElementAdapter}.
	 */
	public static boolean isDeviceInternalPredecessor(PlatformConnectorUnit connCandidate,
			IModelElementAdapter<?> element,
			DefaultDirectedGraph<PlatformConnectorUnit, DefaultEdge> route) {
		// The following rules apply for device internal communication:
		// - Sender Port is on the same device OR
		// - its reference port is one layer "below", e.g. Gateway -> Router AND
		// - the sender connector must have been traversed (natural for path walking a multicast
		// message.
		if(route.containsVertex(connCandidate) && (isConnectorOf(connCandidate, element) ||
				isContainedInResource(connCandidate, element))) {
			return true;
		}
		return false;
	}

	/**
	 * Returns the collection of {@link InputPort}s and {@link OutputPort}s attached to the given
	 * {@link Component}.
	 * 
	 * @param component
	 *            Containing {@link Component}.
	 * @return Collection of attached {@link InputPort}s and {@link OutputPort}s as {@link Port}s.
	 */
	public static Collection<Port> getInputOutputPortsOf(Component component) {
		Collection<Port> ports = new ArrayList<Port>();
		ports.addAll(component.getInputPorts());
		ports.addAll(component.getOutputPorts());
		return ports;
	}

	/**
	 * Returns the collection of {@link Port}s which are connected via {@link Channel}s with the
	 * given {@link Port}
	 * 
	 * @param port
	 *            {@link Port} whose connected {@link Port}s shall be found.
	 * @return Collection of connected {@link Port}s.
	 */
	public static Collection<Port> getConnectedPortsOf(Port port) {
		Collection<Port> connectedPorts = new ArrayList<Port>();
		port.getIncomingChannels().stream().forEach(conn -> connectedPorts.add(conn.getSource()));
		port.getOutgoingChannels().stream().forEach(conn -> connectedPorts.add(conn.getTarget()));
		return connectedPorts;
	}

	/**
	 * Returns the collection of <i>all</i> directly connected {@link Channel}s of the given
	 * {@link Component}.
	 * 
	 * @param component
	 *            Subject {@link Component}.
	 * @return Collection of directly connected {@link Channel}s.
	 */
	public static Collection<Channel> getConnectedChannelsOf(Component component) {
		Collection<Port> containedPorts = getInputOutputPortsOf(component);
		Collection<Channel> connectedChannels = new ArrayList<>();
		containedPorts.stream().forEach(
				p -> p.getIncomingChannels().stream().forEach(ch -> connectedChannels.add(ch)));
		containedPorts.stream().forEach(
				p -> p.getOutgoingChannels().stream().forEach(ch -> connectedChannels.add(ch)));
		return connectedChannels;
	}

	/**
	 * Searches the given {@link ComponentArchitecture} for {@link ComponentDiverseImplRef}s and
	 * returns all referenced {@link Component}s (representing an implementation of the referencing
	 * {@link Component}).
	 * 
	 * @param compArch
	 *            The {@link ComponentArchitecture} to be searched for references.
	 * @return Collection of all {@link Component}s referenced by the
	 *         {@link ComponentDiverseImplRef} {@link IModelElementSpecification} in the given
	 *         {@link ComponentArchitecture} :
	 */
	// public static Collection<Component> getAllReferencedDiverseComponentsOf(
	// ComponentArchitecture compArch) {
	// Collection<Component> referencedComponents = new ArrayList<Component>();
	// List<ComponentDiverseImplRef> diversitySpec =
	// getChildrenWithType(compArch, ComponentDiverseImplRef.class);
	// for(ComponentDiverseImplRef diverseComponentSpec : diversitySpec) {
	// // We assume that the diverse components of one component are located in one and the
	// // same ComponentArchitecture. Hence, it suffices to determine the parent
	// // ComponentArchitecture by the first element in the List.
	// referencedComponents.addAll(diverseComponentSpec.getComponentRef());
	// }
	// return referencedComponents;
	// }

	/**
	 * Returns the first identified <i>Component Pool</i> that is associated with the given
	 * {@link ComponentArchitecture}. This method uses the set of diverse implementations of
	 * {@link ITaskAdapter}s from the input {@link ComponentArchitecture} and returns
	 * its containing {@link ComponentArchitecture}. This method assumes that there exists maximally
	 * one <i>Component Pool</i> per "actual" {@link ComponentArchitecture}.
	 * 
	 * @param mapping
	 *            {@link IMappingEncoding} whose {@link ITaskAdapter}s referenced by
	 *            the {@link IMappingEntry}s are used to identify the <i>Component Pool</i>.
	 * @return The {@link ComponentArchitecture} identified as the <i>Component Pool</i> or
	 *         {@code null} if none is found.
	 */
	public static ComponentArchitecture getFirstComponentPool(
			IMappingEncoding<ITaskAdapter<?>, IResourceAdapter<?>, IMappingEntry<ITaskAdapter<?>, IResourceAdapter<?>>> mapping) {
		for(ITaskAdapter<?> deployableComponent : mapping.getRequesters()) {
			if(deployableComponent.isInstantiation()) {
				// TODO: Avoid casting: Parameterize the IModelElementAdapter to use EObject-derived
				// generics?
				ComponentArchitecture parentCA =
						getParentElement((EObject)deployableComponent.getObject(),
								ComponentArchitecture.class, false);
				if(parentCA != null) {
					return parentCA;
				}
			} else if(deployableComponent.isInstantiationRequired()) {
				for(ITaskAdapter<?> replacementComponent : deployableComponent
						.getReplacementComponents()) {
					// TODO: Avoid casting: Parameterize the IModelElementAdapter to use
					// EObject-derived generics?
					ComponentArchitecture parentCA =
							getParentElement((EObject)replacementComponent.getObject(),
									ComponentArchitecture.class, false);
					if(parentCA != null) {
						return parentCA;
					}
				}
			}
		}
		return null;
	}

	/**
	 * Resets the IDs of the given {@code root} object and its contained elements to 0.
	 * 
	 * @param root
	 *            {@link EObject} for which and for whose contained elements, the id will be set to
	 *            0.
	 */
	public static void resetIds(IIdLabeled root) {
		for(TreeIterator<EObject> iter = root.eAllContents(); iter.hasNext();) {
			EObject obj = iter.next();
			if(obj instanceof IIdLabeled) {
				((IIdLabeled)obj).setId(0);
			}
		}
		root.setId(0);
	}

	/**
	 * Checks whether any parent {@link Component} of the given {@link EObject} is chosen to be
	 * deployable element.
	 * 
	 * @param startElement
	 *            The element whose parents shall be evaluated w.r.t. the deployable property
	 * @param systemModelAdapter
	 *            Abstraction of the input models.
	 * @return True if any parent {@link Component} is chosen to be deployed, false otherwise.
	 * @throws InvalidTimingModelException
	 *             if the deployable element cannot be identified.
	 */
	public static <T extends EObject> boolean isAnyParentComponentDeployable(EObject startElement,
			SystemModelAdapter<?, ?, ?, ?, ?, ?, ?> systemModelAdapter)
			throws InvalidTimingModelException {
		EList<Component> parentElements = getParentsWithType(startElement, Component.class);

		for(Component parentComponent : parentElements) {
			if(systemModelAdapter.isDeployableComponent(parentComponent)) {
				return true;
			}
		}

		return false;
	}

	//
	/**
	 * Returns whether any container {@link Component} of the given {@code startElement}
	 * {@link Component} is a deployable {@link Component}. Therefore the
	 * {@code encEntryModelElemAssoc} map is checked for containing any entry that is associated
	 * with one of the containing {@link Component}s.
	 *
	 * @param startElement
	 *            {@link Component} to check.
	 * @param encEntryModelElemAssoc
	 *            associations between {@link TaskMappingEntry}s and {@link Component}s.
	 * @return Whether any containing {@link Component} is deployed.
	 * @throws InvalidTimingModelException
	 */
	public static <T extends EObject> boolean isAnyParentComponentDeployable(EObject startElement,
			TypedBiMap<Object, EObject> encEntryModelElemAssoc) throws InvalidTimingModelException {
		EList<Component> parentElements = getParentsWithType(startElement, Component.class);

		for(Component parentComponent : parentElements) {
			AbstractTaskMappingEntry entry = (AbstractTaskMappingEntry)encEntryModelElemAssoc
					.inverse().get(parentComponent, AbstractTaskMappingEntry.class);
			if(entry != null) {
				return true;
			}
		}

		return false;
	}
}
